/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package seamcarving.lib;

import java.awt.image.BufferedImage;
import seamcarving.ui.SeamCarving;

/**
 *
 * @author Aleš
 */
public class Retargeting {
    
    private int width, height; // source image dimensions
    private BufferedImage source; // source image
    private int[] rgb; // byte source image
    private BufferedImage sourceY; // Y of source image from YCbCr
    private int[] luma; // byte Y
    private BufferedImage sourceE; // edges of source image
    private int[] edges; // byte edges
    private int[] Redges; // resized edges
    private int[] Rrgb; // resized RGB
    private int[] sums; // downwards search for lowest energy path
    private short[] path; // path for seam backtracking
    private int count; // amount of removed seams
    private int step; // currently removed seams
    private BufferedImage seams; // Y image with seams in red
    private BufferedImage result; // final carved image
    private SeamCarving parentForm; // caller UI

    public Retargeting(BufferedImage input, SeamCarving parent) {
        source = input;
        width = source.getWidth();
        height = source.getHeight();
        parentForm = parent;
        luma = new int[width*height];
        sourceY = new BufferedImage(width, height,BufferedImage.TYPE_INT_RGB);
        int z;
        rgb = source.getRGB(0, 0, width, height, null, 0, width);
        int[] work = new int[width*height];
        for (int i=0; i<width*height; i++) { // converts source to Y from YCbCr
            z = (int)(0.299*(rgb[i] >> 16 & 0xff)+0.587*(rgb[i] >> 8 & 0xff)+0.114*(rgb[i] & 0xff));
            luma[i] = z;
            work[i] = 0xff000000 | z << 16 | z << 8 | z;
        }
        sourceY.setRGB(0, 0, width, height, work, 0, width);
    }
    public BufferedImage getLuma() {
        return sourceY;
    }
    public BufferedImage getEdges() {
        return sourceE;
    }
    public BufferedImage getResult() {
        return result;
    }
    // finds edges by Gauss and Sobel convolutions
    public BufferedImage FindEdges() {
        edges = new int[width*height];
        Redges = new int[width*height];
        sourceE = new BufferedImage(width, height,BufferedImage.TYPE_INT_RGB);
        System.arraycopy(luma, 0, edges, 0, width*height);
        for (int i=2*width+2; i<width*(height-2)-2; i++) {
            edges[i] = (2*luma[i-2*width-2]+4*luma[i-2*width-1]+5*luma[i-2*width]+4*luma[i-2*width+1]+2*luma[i-2*width+2]
                    +4*luma[i-width-2]+9*luma[i-width-1]+12*luma[i-width]+9*luma[i-width+1]+4*luma[i-width+2]
                    +5*luma[i-2]+12*luma[i-1]+15*luma[i]+12*luma[i+1]+5*luma[i+2]
                    +4*luma[i+width-2]+9*luma[i+width-1]+12*luma[i+width]+9*luma[i+width+1]+4*luma[i+width+2]
                    +2*luma[i+2*width-2]+4*luma[i+2*width-1]+5*luma[i+2*width]+4*luma[i+2*width+1]+2*luma[i+2*width+2])/159;
            /*
            edges[i] = (luma[i-2*width-2]+luma[i-2*width-1]+luma[i-2*width]+luma[i-2*width+1]+luma[i-2*width+2]
                    +luma[i-width-2]+luma[i-width-1]+luma[i-width]+luma[i-width+1]+luma[i-width+2]
                    +luma[i-2]+luma[i-1]+luma[i]+luma[i+1]+luma[i+2]
                    +luma[i+width-2]+luma[i+width-1]+luma[i+width]+luma[i+width+1]+luma[i+width+2]
                    +luma[i+2*width-2]+luma[i+2*width-1]+luma[i+2*width]+luma[i+2*width+1]+luma[i+2*width+2])/25;
            */
        }
        int[] dest = new int[width*height];
        int x, y;
        double x2, y2;
        int z;
        int j;
        for (j=0; j<width; j++) { // first row
            x2 = Math.pow(edges[j], 2);
            y2 = Math.pow(edges[j], 2);
            z = (int) Math.sqrt((x2+y2)/2.f);
            dest[j] = 0xff000000 | z << 16 | z << 8 | z;
            edges[j] = z;
        }
        for (j=width*(height-1); j<width*height; j++) { // last row
            x2 = Math.pow(edges[j], 2);
            y2 = Math.pow(edges[j], 2);
            z = (int) Math.sqrt((x2+y2)/2.f);
            dest[j] = 0xff000000 | z << 16 | z << 8 | z;
            edges[j] = z;
        }
        for (j=1; j<height-1; j++) { // first column
            x2 = Math.pow(edges[j*width], 2);
            y2 = Math.pow(edges[j*width], 2);
            z = (int) Math.sqrt((x2+y2)/2.f);
            dest[j*width] = 0xff000000 | z << 16 | z << 8 | z;
            edges[j*width] = z;
        }
        for (j=1; j<height-1; j++) { // last column
            x2 = Math.pow(edges[(j+1)*width-2], 2);
            y2 = Math.pow(edges[(j+1)*width-2], 2);
            z = (int) Math.sqrt((x2+y2)/2.f);
            dest[(j+1)*width-2] = 0xff000000 | z << 16 | z << 8 | z;
            edges[(j+1)*width-2] = z;
        }
        for (j=width+1; j<width*(height-1)-1; j++) { // inside
            x = -luma[j-width-1]+luma[j-width+1]-2*luma[j-1]+2*luma[j+1]-luma[j+width-1]+luma[j+width+1];
            y = luma[j-width-1]+2*luma[j-width]+luma[j-width+1]-luma[j+width-1]-2*luma[j+width]-luma[j+width+1];
            z = (int) Math.sqrt((Math.pow(x, 2)+Math.pow(y, 2))/2.f);
            dest[j] = 0xff000000 | z << 16 | z << 8 | z;
            edges[j] = z;
        }
        sourceE.setRGB(0, 0, width, height, dest, 0, width);
        return sourceE;
    }
    public void setCount( int N) {
            count = N;
        }
    // creates new thread with retargeting and starts it
    public void Run() {
        Task task = new Task();
        task.start();
    }
    private class Task extends Thread {
        // recalculates sums and path
        private void Recalculate() {
            int W = width-step;
            sums = new int[W*height];
            path = new short[W*(height-1)];
            System.arraycopy(Redges, 0, sums, 0, W);
            short k;
            int min;
            for (int i=0; i<W*(height-1); i++) {
                if ( i % W == 0) {
                    if (sums[i] < sums[i+1]) {
                        sums[i+W] = Redges[i+W]+sums[i];
                        path[i] = 0;
                        continue;
                    }
                    sums[i+W] = Redges[i+W]+sums[i];
                    path[i] = 1;
                    continue;
                }
                if ( i % W == W-1) {
                    if (sums[i-1] < sums[i]) {
                        sums[i+W] = Redges[i+W]+sums[i-1];
                        path[i] = -1;
                        continue;
                    }
                    sums[i+W] = Redges[i+W]+sums[i];
                    path[i] = 0;
                    continue;
                }
                k = -1;
                min = sums[i-1];
                if (sums[i] < min) {
                    k = 0;
                    min = sums[i];
                }
                if (sums[i+1] < min) {
                    k = 1;
                    min = sums[i+1];
                }
                sums[i+W] = Redges[i+W]+min;
                path[i] = k;
            }
        }
        // finds, removes and draws seams
        @Override
        public void run() {
            seams = new BufferedImage(width, height,BufferedImage.TYPE_INT_RGB);
            result = new BufferedImage(width-count, height,BufferedImage.TYPE_INT_RGB);
            seams.getGraphics().drawImage(sourceE, 0, 0, null);
            Rrgb = new int[width*height];
            System.arraycopy(rgb, 0, Rrgb, 0, width*height);
            Redges = new int[width*height];
            System.arraycopy(edges, 0, Redges, 0, width*height);
            int seam[] = new int[height];
            int points[] = new int[height*count];
            int[] tmpRGB;
            int[] tmpEdges;
            for (int l=0; l<count; l++) {
                step = l;
                int W = width-l;
                Recalculate();
                tmpRGB = new int[W*height];
                tmpEdges = new int[W*height];
                System.arraycopy(Redges, 0, tmpEdges, 0, W*height);
                System.arraycopy(Rrgb, 0, tmpRGB, 0, W*height);
                Redges = new int[W*height];
                Rrgb = new int[W*height];
                int min = sums[W*(height-1)];
                int mi = 0;
                for (int i=0; i<W; i++) {
                    if (sums[W*(height-1)+i] < min) {
                        min = sums[W*(height-1)+i];
                        mi = i;
                    }
                }
                seam[height-1] = W*(height-1)+mi;
                points[(l+1)*height-1] = mi;
                int e = 0;
                for (int i=0; i<l; i++) {
                    if (points[(i+1)*height-1] <= mi)
                        e++;
                }
                seams.setRGB(mi+e, height-1, 0xffff0000);
                e = 0;
                int d = 0;
                for (int j=height-2; j>=0; j--) {
                    d += path[j*W+mi+d];
                    points[(l+1)*height-j-1] = mi+d;
                    for (int i=0; i<l; i++) {
                        if (points[(i+1)*height-j-1] <= mi+d)
                            e++;
                    }
                    seams.setRGB(mi+d+e, j, 0xffff0000);
                    e = 0;
                    seam[j] = j*W+mi+d;
                }
                int omit = 0;
                for (int k=0; k<(W-1)*height; k++) { // egde smoothing
                    if (omit < height && k+omit == seam[omit]) {
                        omit++;
                        Rrgb[k] = combine(tmpRGB[k+omit], Rrgb[k-1]);
                    } else
                        Rrgb[k] = tmpRGB[k+omit];
                    Redges[k] = tmpEdges[k+omit];
                }
            }
            result.setRGB(0, 0, width-count, height, Rrgb, 0, width-count);
            parentForm.getCanvasSeam().buffImage(seams);
            parentForm.getCanvasRslt().buffImage(result);
            parentForm.unlockGUI();
        }
    }
    
    private int combine( int a, int b) { // seam smoothing
        int R, G, B;
        R = ((a >> 16 & 0xff) + (b >> 16 & 0xff))/2;
        G = ((a >> 8 & 0xff) + (b >> 8 & 0xff))/2;
        B = ((a & 0xff) + (b & 0xff))/2;
        return 0xff000000 | R << 16 | G << 8 | B;
    }
}